"""
This file defines all functionality for Structural Causal Models.
"""

from typing import Tuple, Any, Callable, Dict, List
import networkx as nx
import matplotlib.pyplot as plt
import random


class StructuralCausalModel:
    """
    Class defining a Structural Causal Model. We define an SCM as a tuple
    .. math::
        \mathcal{M} = (\mathcal{X}, \mathcal{F}, \mathcal{U}, \mathcal{P})
    where :math:`\mathcal{X}` is the set of endogenous variables, :math:`\mathcal{F}` is the set of assignment
    functions, :math:`\mathcal{U}` the set of exogenous variables, and :math:`\mathcal{P}` is the set of probability
    distributions associated with the exogenous variables.
    """

    endogenous_vars: Dict[str, Any]
    """Dictionary storing the value of each endogenoous variable."""
    functions: Dict[str, Tuple[Callable, dict]]
    """Functional assignments of the endogenous variables. for each endogenous variables, the functional assignments are
    stored as well as a dictionary mapping the parameters of the callable (key) to the name of the causes (values)."""
    exogenous_vars: Dict[str, Any]
    """Dictionary storing the values of the exogenous variables."""
    exogenous_distributions: Dict[str, Tuple[Callable, dict]]
    """Dictionary storing the distribution for each exogenous variable. The values of the dict contain the callable 
    representing the distribution and it's kwargs as a tuple."""
    saved_functions: Dict[str, Tuple[Callable, dict]]
    """Contains a backup of the function of each endogenous variable to be able to restore them after intervention."""
    def __init__(self):
        self.endogenous_vars = {}
        self.exogenous_vars = {}
        self.functions = {}
        self.exogenous_distributions = {}
        self.saved_functions = {}

    def add_endogenous_var(self, name: str, function: Callable, param_varnames: dict):
        """
        Adds an endogenous variable to the SCM.

        :param name: name of the endogenous variable
        :param function: callable that returns a value given the causes of the endogenous variables.
        :param param_varnames: dict that maps names of the parameters in the function to the name of the parent node.

        Example:
        >>> scm.add_endogenous_var('A', lambda noise: noise + 5, {'noise': 'U'})
        """
        # all names are uppercase
        name = name.upper()
        # ensure unique names
        assert name not in self.exogenous_vars.keys(), 'Variable already exists'
        assert name not in self.endogenous_vars.keys(), 'Variable already exists in endogenous vars'

        self.endogenous_vars[name] = None
        self.functions[name] = (function, param_varnames)

    def add_endogenous_vars(self, vars: List[Tuple[str, Callable, dict]]):
        """
        Adds a list of endogenous variables to the SCM.

        :param vars: list of endogenous variables definitions as defined in the 'add_endogenous_var' function.
        """
        [self.add_endogenous_var(v[0], v[1], v[2]) for v in vars]

    def add_exogenous_var(self, name: str, distribution: Callable, distribution_kwargs: dict):
        """
        Add an exogenous variable to the SCM.

        :param name: name of the exogenous variable.
        :param distribution: distribution of the exogenous variable.
        :param distribution_kwargs: kwargs for the distribution

        Example:
        >>> scm.add_exogenous_var('U', random.randint, {'a': 3, 'b': 8})
        """
        # all names are uppercase
        name = name.upper()
        # ensure unique names
        assert name not in self.exogenous_vars.keys(), 'Variable already exists'
        assert name not in self.endogenous_vars.keys(), 'Variable already exists in endogenous vars'

        self.exogenous_vars[name] = None
        self.exogenous_distributions[name] = (distribution, distribution_kwargs)

    def add_exogenous_vars(self, vars: List[Tuple[str, Any, Callable, dict]]):
        """
        Adds a list of exogenous variables to the SCM.

        :param vars: list of exogenous variables definitions as defined in the 'add_exogenous_var' function.
        """
        [self.add_exogenous_var(v[0], v[1], v[2]) for v in vars]

    def get_next_sample(self) -> Tuple[Dict, Dict]:
        """
        Generates an ancestral sample from the joint distribution of endogenous variables by generating a sample of the
        exogenous variables
        .. math::
            u \sim P(\mathcal{U})
        and determining the value of variables :math:`\mathcal{X}` by applying the functions in :math:`\mathcal{F}`
        according to the topological ordering in the causal graph resulting in a sample of the endogenous variables
        .. math::
            x \sim P(\mathcal{X})


        :return: a sample of endogenous :math:`x` and exogenous variables :math:`u`.
        """
        random.seed()
        # update exogenous vars
        for key in self.exogenous_vars:
            dist = self.exogenous_distributions[key]
            res = dist[0](**dist[1])
            self.exogenous_vars[key] = res

        # update endogenous vars
        structure_model = self.create_graph()
        node_order = [n for n in nx.topological_sort(structure_model)]
        # remove exogenous vars since they are root nodes and already have a value
        [node_order.remove(n) for n in self.exogenous_vars]
        # propagate causal effects along the topological ordering
        for node in node_order:
            # get the values for the parameters needed in the functions
            params = {}
            for param in self.functions[node][1]:  # parameters of functions
                if self.functions[node][1][param] in self.endogenous_vars.keys():
                    params[param] = self.endogenous_vars[self.functions[node][1][param]]
                else:
                    params[param] = self.exogenous_vars[self.functions[node][1][param]]

            # Update variable according to its function and parameters
            self.endogenous_vars[node] = self.functions[node][0](**params)
        return dict(self.endogenous_vars), dict(self.exogenous_vars)

    def do_interventions(self, interventions: List[Tuple[str, Tuple[Callable, dict]]]):
        """
        Replaces the functions of the scm with the given interventions per endogenous variable. E.g. the intervention
        :math:`do(X_0 = 5, X_1 = X_0+1)` can be implemented with
        >>> scm.do_interventions([("X0", (lambda: 5, {})), ("X1", (lambda x0: x0+1, {'X0':'x0'})])

        :param interventions: List of tuples where every tuple contains the name of the variable to intervene on and a
        callable that represents the new causal function for this variable given its parents as a dict. The dict maps
        the parameters of the callable (key) to the names of the parents in the SCM (value).
        """
        random.seed()
        self.saved_functions = {}
        for interv in interventions:
            if interv:  # interv not None
                self.saved_functions[interv[0]] = self.functions[interv[0]]
                self.functions[interv[0]] = interv[1]

    def undo_interventions(self):
        """
        Restores all functional relations that were deleted in the previous call of `do_interventions`.
        """
        for key, value in self.saved_functions.items():
            self.functions[key] = value
        self.saved_functions.clear()

    def get_intervention_targets(self) -> List[str]:
        """
        Returns a list containing the names of the variables that are currently being intervened on.

        :return: List of intervention targets.
        """
        return list(self.saved_functions.keys())

    def create_graph(self) -> nx.DiGraph:
        """
        Returns the DAG that corresponds to the functional structure of this SCM.

        :return: A causal graph.
        """
        graph = nx.DiGraph()

        # create nodes
        [graph.add_node(var.upper(), type='endo') for var in self.endogenous_vars]
        [graph.add_node(var.upper(), type='exo') for var in self.exogenous_vars]

        for var in self.functions:
            for parent in self.functions[var][1].values():
                if parent.lower() in self.endogenous_vars or parent.upper() in self.endogenous_vars\
                        or parent.lower() in self.exogenous_vars or parent.upper() in self.exogenous_vars:
                    graph.add_edge(parent.upper(), var.upper())

        return graph

    def draw_graph(self):
        """
        Draws the causal graph.
        """
        graph = self.create_graph()
        values = dict(self.endogenous_vars), dict(self.exogenous_vars)
        values = dict(values[0], **values[1])
        nx.draw(graph, arrowsize=20, with_labels=True, node_size=3000, font_size=10,
                labels={key: str(key) + ':\n' + str(values[key]) for key in values}, pos=nx.planar_layout(graph))
        plt.show()




"""
# Overview
The [CausalPlayground](https://github.com/sa-and/CausalPlayground) library serves as a tool for causality research,
focusing on the interactive exploration of structural
causal models (SCMs). It provides extensive functionality for creating, manipulating and sampling SCMs, seamlessly
integrating them with the Gymnasium framework. Users have complete control over SCMs, enabling precise manipulation and
interaction with causal mechanisms. Additionally, CausalPlayground offers a range of useful helper functions for generating
diverse instances of SCMs and DAGs, facilitating quantitative experimentation and evaluation. Notably, the library is
designed for easy integration with reinforcement learning methods, enhancing its utility in active inference and
learning settings. This documentation presents the complete API documentation and a quickstart guide [here](https://sa-and.github.io/CausalPlayground/).

# Installation guide (TODO)
Either clone and 'pip install -r requirements.txt'

or 'pip install CausalPlayground'

# Structural Causal Models (SCM)
SCMs are a powerful model to express a data-generating process that is governed by clear causal relations represented
as functions. More formally, an SCM models the causal relationship between endogenous variables, describing the main
random variables of intrerest, and exogenous variables that, roughly speaking, model the context of the process. The
causal effects are described as functional relations between variables, and the exogenous variables are modeled through
distributions. Taken together, an SCM defines the joint probability distribution over its variables. CausalPlayground
facilitates the creation, manipulation, and sampling of SCM as described below.

## Defining an SCM
Define the SCM $\mathcal{M} = (\mathcal{X}, \mathcal{U}, \mathcal{F}, \mathcal{P})$ with
$\mathcal{X}= \\{ A, Effect \\}$, $\mathcal{U}=\\{U\\}$, $\mathcal{P}=\\{Uniform(3, 8)\\}$, and
$\mathcal{F} = \begin{cases}A\leftarrow 5+U\\\Effect \leftarrow A*2\end{cases}$

```Python
>>> scm = StructuralCausalModel()
>>> scm.add_endogenous_var('A', lambda noise: noise+5, {'noise': 'U'})
>>> scm.add_exogenous_var('U', random.randint, {'a': 3, 'b': 8})
>>> scm.add_endogenous_var('Effect', lambda x: x*2, {'x': 'A'})
```

The functions can be defined as lambda functions. Alternatively, any callable function is admissible. The given
dictionary maps the parameters in the callable function to the names of the variables. Note that the variable names will
always be converted to upper-case.

## Sampling the SCM
```Python
>>> x, u = scm.get_next_sample()
>>> x
{'A': 10, 'EFFECT': 20}
>>> u
{'U': 5}
```

## Intervening in an SCM
To perform an intervention $do(X_0 \leftarrow 5, X_1 \leftarrow X_0+1)$, the following can be implemented:
```Python
    >>> scm.do_interventions([("X0", (lambda: 5, {})),
                              ("X1", (lambda x0: x0+1, {'x0':'X0'})])
```
As before, the new function describing the causal relations of a node to its parents, is represented by anny callable
function and the dictionary maps the parameters of that function to the names of the variables. Calling
```Python
    >>> scm.undo_interventions()
```
will return the SCM to its original state with $do(\emptyset)$.

# SCM Environments
## Defining an SCM Environment
When building agents, that interact with an SCM you can use the `CausalPlayground.SCMEnvironment` class.
```Python
>>> from CausalPlayground import SCMEnvironment
>>> env = SCMEnvironment(scm, possible_interventions=[("X0", (lambda: 5, {})), ("X1", (lambda x0: x0+1, {'x0':'X0'}))])
```
The actions in this environment correspond to the interventions defined in the `possible_interventions`.

Calling the `step(action)` function applies the interventions defined in `action` (see explaination below), samples the
intervened SCM, determines the new observation, termination flag, truncated flag and reward, and, finally undoes the
interventions.

To enable multiple interventions simultaneously, a list of interventions can be passed as action. The elements in this
list, correspond to the discrete index of that intervention in the list `possible_interventions`. For example,
simultaneously performing both possible interventions in one step can be done as follows:
```Python
>>> env.step([1, 2])
```
whereas performing a step without intervention can be done by either invoking the first (empty) intervention, or no
intervention as follows:
```Python
>>> env.step([0])
>>> env.step([])
```

## Generating data from the SCM
In its basic implementation, `SCMEnvironment` always returns the current values of the variables as observation and 0
reward regardless of the action. Furthermore, the `terminated` and `truncated` flag are always False, leading to an
episode never terminating.

Consider the following SCM with given possible interventions:
```Python
>>> scm = StructuralCausalModel()
>>> scm.add_endogenous_var('A', lambda noise: noise + 5, {'noise': 'U'})
>>> scm.add_exogenous_var('U', random.randint, {'a': 3, 'b': 8})
>>> scm.add_endogenous_var('EFFECT', lambda x: x * 2, {'x': 'A'})
```
The basic implementation can be used to generate samples (e.g. 1000) from an SCM either with no interventions
(observational data):
```Python
>>> for i in range(1000):
>>>     env.step([0])
```
with some fixed interventions:
```Python
>>> for i in range(1000):
>>>     env.step([1, 2])
```
or even with random interventions:
```Python
>>> for i in range(1000):
>>>     env.step(env.action_space.sample())
```

The collected data can be found in `env.samples_so_far` and the intervention targets in `env.targets_so_far`. These
buffers can be flushed on demand with `clear_samples()` and `clear_intervention_targets()`, respectively.

## Extending SCMEnvironment
While the basic implementation is useful for collecting data generated by an SCM, it holds potential for more sophisticated applications to
reinforcement learning by inheriting from `SCMEnvironment`. To adapt the environment to a specific task, simply override
the neccessary functions determining observation, reward, truncated, and terminated. (Note that the interventions in a
step are undone *after* these quantities are determined.)

Here an example for an environment that rewards interventions:
```Python
>>> from CausalPlayground import SCMEnvironment

>>> class NoIntervEnv(SCMEnvironment):
        def __init__(self, scm, possible_interventions):
            super(NoIntervEnv, self).__init__(scm, possible_interventions)

        def determine_reward(self):
            if len(self.scm.get_intervention_targets()) > 0:
                return 1
            else:
                return 0
>>> env = NoIntervEnv(scm, possible_interventions=[("A", (lambda: 5, {})), ("EFFECT", (lambda x0: x0+1, {'x0': 'A'}))])
```

# Generators
In some scenarios it might be useful to automatically generate graphs and SCMs. To this end, we implemented some helper
classes.

## Graph generation
You can use the `CausalPlayground.CausalGraphGenerator` class to quickly generate directed acyclic graphs. To generate a random
graph with 7 endogenous and 5 exogenous variables, you can run the following:
```Python
>>> from CausalPlayground import CausalGraphGenerator
>>> gen = CausalGraphGenerator(7, 5)
>>> graph = gen.generate_random_graph()
```
This creates graphs with no exogenous confounders. For allowing exogenous confounders, instantiate the generator as
```Python
>>> gen = CausalGraphGenerator(7, 5, allow_exo_confounders=True)
```

With the `CausalPlayground.CausalGraphSetGenerator` class, you can generate many DAGs at once. Additionally, this class makes
sure that each graph is contained only ones in the list of generated graphs.
```Python
>>> from CausalPlayground import CausalGraphSetGenerator
>>> gen = CausalGraphSetGenerator(7, 5)
>>> gen.generate(1000)
```
generates 1000 distinct DAGs with 7 endogenous nodes and 5 exogenous nodes that can be accessed through `gen.graphs`.

## SCM generation
The `CausalPlayground.SCMGenerator` class also let's you generate SCMs automatically, for example with linear additive
causal relations and a random causal structure.
```Python
>>> from tests.functions import f_linear
>>> from CausalPlayground import SCMGenerator

>>> gen = SCMGenerator(all_functions={'linear': f_linear})
>>> scm_unconfounded = gen.create_random(possible_functions=["linear"], n_endo=5, n_exo=4,
                                     exo_distribution=random.random, exo_distribution_kwargs={},
                                     allow_exo_confounders=False)[0]
>>> scm_confounded = gen.create_random(possible_functions=["linear"], n_endo=5, n_exo=7,
                                   exo_distribution=random.random, exo_distribution_kwargs={},
                                   allow_exo_confounders=True)[0]
```
To generate an SCM based on a given causal structure:
```Python
>>> generator.create_scm_from_graph(graph=GRAPH, possible_functions=['linear'],
                                exo_distribution=random.randint, exo_distribution_kwargs={'a': 2, 'b': 5})
```

`./envs/generation/functions.py` defines the functions that can be used when defining your SCM. You can either use the
pre-defined functions, or define the function that is tailored to your need.

The interface to define your own data-generating function is as follows: define an (outer) python function that takes as
input a list of strings (the parent's variable names) and outputs the (inner) data-generation function. Giving the inner
function **kwargs as parameters, allows you to access the parent variable's values. The remaining behaviour of the outer
and inner function can be arbitrarily defined. An example is given below:

```Python
def f_linear(parents: List[str]):
    weights = {p: random.uniform(0.5, 2.0) for p in parents}
    default_value = 0.0

    def f(**kwargs):
        if len(kwargs) == 0:
            mu = default_value
        else:
            mu = 0.0

        for p in parents:
            mu += weights[p] * kwargs[p]
        return mu

    return f
```
This function can then be passed to the generator as described above.

"""
from .scm import StructuralCausalModel
from .generators import SCMGenerator, CausalGraphGenerator, CausalGraphSetGenerator
from .scm_environment import SCMEnvironment



import unittest
import numpy as np
from CausalPlayground.generators import CausalGraphGenerator, CausalGraphSetGenerator
import networkx as nx
import os


class TestGraphGeneration(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(TestGraphGeneration, self).__init__(*args, **kwargs)
        self.sizes = [3, 4, 10, 15]
        self.generators = {n: CausalGraphGenerator(n_endo=n, n_exo=n-3) for n in self.sizes}
        self.unconfounded_test_graph = CausalGraphGenerator(n_endo=5,
                                                            n_exo=10,
                                                            allow_exo_confounders=False).generate_random_graph()[0]
        self.confounded_test_graph = CausalGraphGenerator(n_endo=10,
                                                          n_exo=100,
                                                          allow_exo_confounders=True).generate_random_graph()[0]
        
    def test_type(self):
        self.assertEqual(type(self.unconfounded_test_graph), nx.DiGraph)

    def test_n_nodes(self):
        [self.assertEqual(len(self.generators[n].generate_random_graph()[0].nodes), 2*n-3) for n in self.sizes]

    def test_acyclyc(self):
        [self.assertTrue(nx.is_directed_acyclic_graph(self.generators[n].generate_random_graph()[0])) for n in self.sizes]

    def test_endo_exo(self):
        for node in self.unconfounded_test_graph.nodes:
            self.assertIn(self.confounded_test_graph.nodes[node]['type'], ['endo', 'exo'])

    def test_exo_root(self):
        '''
        Test whether exogenous variables are roots of the graph
        '''
        for graph in [self.confounded_test_graph, self.unconfounded_test_graph]:
            for node in graph.nodes:
                if graph.nodes[node]['type'] == 'exo':
                    self.assertIn(len([i for i in graph.predecessors(node)]), [0, 1])

    def test_confounding(self):
        for node in self.unconfounded_test_graph.nodes:
            if self.unconfounded_test_graph.nodes[node]['type'] == 'exo':
                self.assertIn(len([n for n in self.unconfounded_test_graph.successors(node)]), [0, 1])

    def test_n_generated_graphs(self):
        set_generator = CausalGraphSetGenerator(n_endo=5, n_exo=10, allow_exo_confounders=False)
        set_generator.generate(100)
        self.assertEqual(len(set_generator.graphs), 99)
        
    def test_seeding(self):
        set_generator1 = CausalGraphSetGenerator(n_endo=5, n_exo=5, allow_exo_confounders=False, seed=1)
        set_generator2 = CausalGraphSetGenerator(n_endo=5, n_exo=5, allow_exo_confounders=False, seed=50)
        set_generator3 = CausalGraphSetGenerator(n_endo=5, n_exo=5, allow_exo_confounders=False, seed=1)
        set_generator1.generate(50)
        set_generator2.generate(50)
        set_generator3.generate(50)
        
        for i in range(49):
            self.assertTrue(nx.is_isomorphic(set_generator1.graphs[i], set_generator3.graphs[i]))
            self.assertFalse(nx.is_isomorphic(set_generator1.graphs[i], set_generator2.graphs[i]))  # holds just for
            # these specific seeds and 50 samples

    def test_graphset_save_load(self):
        graph_set = CausalGraphSetGenerator(n_endo=5, n_exo=10, allow_exo_confounders=False)
        graph_set.generate(20)
        graph_set.save('delme.pkl')
        graph_set.load('delme.pkl')
        self.assertTrue(len(graph_set.graphs), 20)
        os.remove('delme.pkl')
        
    def test_unique_graph_set(self):
        graph_set = CausalGraphSetGenerator(n_endo=4, n_exo=0, allow_exo_confounders=False)
        graph_set.generate(300)
        edges = [list(g.edges()) for g in graph_set.graphs]
        self.assertTrue(all([len(e) == len(set(e)) for e in edges]))
        graph_set.generate(300, method='ER', p=0.3)
        edges = [list(g.edges()) for g in graph_set.graphs]
        self.assertTrue(all([len(e) == len(set(e)) for e in edges]))

    def test_save_load_graph_set(self):
        graph_set = CausalGraphSetGenerator(n_endo=4, n_exo=0, allow_exo_confounders=False)
        graph_set.generate(300)
        graph_set.save('delme.pkl')
        graph_set.load('delme.pkl')
        np.array([1, 2]).dump('delme.pkl')
        with self.assertRaises(TypeError):
            graph_set.load('delme.pkl')
        

if __name__ == '__main__':
    unittest.main()



import unittest
from CausalPlayground import SCMEnvironment
from CausalPlayground import StructuralCausalModel
import random


class TestSCMEnvironment(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(TestSCMEnvironment, self).__init__(*args, **kwargs)
        self.test_scm = StructuralCausalModel()
        self.test_scm.add_exogenous_var('U', random.randint, {'a': 3, 'b': 8})
        self.test_scm.add_endogenous_var('A', lambda noise: noise + 5, {'noise': 'U'})
        self.test_scm.add_endogenous_var('B', lambda: 1, {})
        self.test_scm.add_endogenous_var('Effect', lambda x, y: x * 2 + y, {'x': 'A', 'y': 'B'})

        self.env = SCMEnvironment(scm=self.test_scm,
                                  possible_interventions=[("A", (lambda: 5, {})),
                                                          ("B", (lambda x0: x0 + 1, {'x0': 'A'}))],
                                  render_mode='human')

    def test_possible_interventions(self):
        self.assertTrue(self.env.action_space.feature_space.n == 3)
        mapped_sample = [self.env.get_intervention_from_action(a) for a in self.env.action_space.sample((5, None))]
        self.assertTrue(all([s in self.env.possible_interventions for s in mapped_sample]))

    def test_steps(self):
        for step in range(5):
            action = self.env.action_space.sample()
            print(action)
            self.env.step(action)

    def test_buffers(self):
        """Test wether the buffers for the histories of samples are correct"""
        self.env.update_values_from_scm_sample()
        self.env.clear_samples()
        self.env.clear_intervention_targets()
        self.assertTrue(len(self.env.samples_so_far) == 0)
        self.assertTrue(len(self.env.targets_so_far) == 0)

        self.env.scm.do_interventions([("A", (lambda: 5, {})), ("B", (lambda x0: x0 + 1, {'x0': 'A'}))])
        self.env.update_values_from_scm_sample()
        self.assertTrue(len(self.env.samples_so_far) == 1)
        self.assertTrue(len(self.env.targets_so_far) == 1)
        self.assertDictEqual(self.env.targets_so_far.iloc[-1].to_dict(), {'A': 1, 'B': 1, 'EFFECT': 0})

        self.env.scm.undo_interventions()
        self.env.update_values_from_scm_sample()
        self.assertDictEqual(self.env.targets_so_far.iloc[-1].to_dict(), {'A': 0, 'B': 0, 'EFFECT': 0})

    def test_rendering(self):
        for _ in range(3):
            self.env.scm.get_next_sample()
            self.env.render()
            self.env.render_mode = 'terminal'
            self.env.render()

    def test_add_intervention(self):
        self.assertEquals(len(self.env.possible_interventions), 3)
        self.env.add_possible_intervention(('EFFECT', (lambda x: 3*x, {'x': 'U'})))
        self.assertEquals(len(self.env.possible_interventions), 4)
        with self.assertRaises(AssertionError):
            self.env.add_possible_intervention(('X', (lambda x: 3 * x, {'x': 'U'})))  # X is not part of the SCM
        self.assertTrue(any([i[0] == 'EFFECT' for i in self.env.possible_interventions[1:]]))  # Ignor None intervention


if __name__ == '__main__':
    unittest.main()



"""Defines the classes for the Gym environments that use an SCM"""

from typing import List, Callable, Tuple, NoReturn, Dict, Any, SupportsFloat
import networkx as nx
from gymnasium import Env
import numpy as np
from CausalPlayground import StructuralCausalModel
from pandas import DataFrame
from gymnasium import spaces


class SCMEnvironment(Env):
    """
    Defines a GYM environment over an SCM. Inherited from gymnasium.Env, please refer to
    https://gymnasium.farama.org/api/env/#gymnasium.Env for further documentation.
    """

    render_mode: str | None = None
    """Mode in which to render the environment. E.g.: 'human', 'terminal', None"""
    steps_this_episode: int
    """Number of steps that have been performed in the current episode."""
    current_action_indicies: List[int]
    """List of indices in the discrete action space that indicate the intervention in 'possible_interventions'."""
    current_interventions: List[Tuple[str, Tuple[Callable, Dict]]]
    """List of indices in the interventions corresponding to the current action indicies."""
    possible_interventions: List[Tuple[str, Tuple[Callable, Dict]]]
    """List of all possible interventions. Each intervention is represented by a tuple with the intervention target
    as first element and the callable + a dictionary mapping the parameters of the callable to the variables names as
    second element. The empty intervention $do(\emptyset)$ is always part of the possible interventions."""
    scm: StructuralCausalModel
    """SCM object that that represents the the data-generating process of the environment."""
    samples_so_far: DataFrame
    """Data frame containing all the samples that have been collected by this environment. The rows correspond to
    the targets in 'targets_so_far'."""
    targets_so_far: DataFrame
    """Data frame containing all intervention targets of the environment so far. The targets are encoded one-hot where
    1 means that the variable has been intervened on and 0 means it has not been intervened on. The rows correspond to
    the samples in 'samples_so_far'."""
    observation_space: spaces.Box
    """Observation space of the environment for a vector containing the values of the variables of the SCM."""
    action_space: spaces.Sequence
    """Action space of the environment for a list of discrete actions representing the indicies of the interventions 
    that are performed on the SCM"""

    def __init__(self, scm: StructuralCausalModel,
                 possible_interventions: List[Tuple[str, Tuple[Callable, Dict]]],
                 render_mode: str | None = None,
                 seed: int | None = None) -> None:
        """
        Constructor for SCM Gymnasium environment.

        :param scm: The SCM for the data-generating process of the environment
        :param possible_interventions: List of tuples where every tuple contains the name of the variable to intervene
        on and a callable that represents the new causal function for this variable given its parents as a dict. The
        dict maps the parameters of the callable (key) to the names of the parents in the SCM (value).
        :param render_mode: what way to render the environment. 'human' plots the causal graph + current values,
        'terminal' prints the current values of the variables.
        :param seed: seed for the random number generator.
        """
        super(SCMEnvironment, self).__init__()
        self.render_mode = render_mode
        self.steps_this_episode = 0
        self.current_action_indicies = []
        self.current_interventions = []
        self.scm = scm
        self.possible_interventions = [None]
        [self.add_possible_intervention(i) for i in possible_interventions]

        # create sample buffer that collects all samples
        self.samples_so_far = DataFrame(columns=list(self.scm.endogenous_vars.keys()) +
                                                list(self.scm.exogenous_vars.keys()))
        # create buffer that collects all intervention targets. 1 if intervened 0 if not
        self.targets_so_far = DataFrame(columns=list(self.scm.endogenous_vars.keys()))

        self.endogenous_vars = self.exogenous_vars = None

        self.observation_space = spaces.Box(low=-np.inf, high=np.inf,
                                            shape=(len(self.scm.endogenous_vars)+len(self.scm.exogenous_vars),))
        self.action_space = spaces.Sequence(spaces.Discrete(len(self.possible_interventions)))

        self.reset(seed=seed)

    def step(self, action: list[int]) -> tuple[np.ndarray, SupportsFloat, bool, bool, dict[str, Any]]:
        """
        Performs an action in the environment and returns the effects. This corresponds to simultaneously applying a
        list of interventions to the SCM. New values are sampled from the SCM and reward, terminated and truncated flag
        are returned.

        :param action: List of intervention indicies according to the possible_interventions defined in the constructor
        :return: see https://gymnasium.farama.org/api/env/#gymnasium.Env
        """
        # Map action indicies to interventions
        self.current_interventions = [self.get_intervention_from_action(a) for a in action]

        # apply intervention
        self.scm.do_interventions(self.current_interventions)

        # sample the environment's SCM
        self.update_values_from_scm_sample()

        self.render()

        self.steps_this_episode += 1

        # task-specific quantities
        observation = self.get_observation()
        reward = self.determine_reward()
        truncated = self.determine_truncated()
        terminated = self.determine_terminated()

        self.scm.undo_interventions()

        if terminated or truncated:
            # reset environment if episode is done
            self.reset()

        return observation, reward, terminated, truncated, {}

    def reset(self, seed: int | None = None, options: dict[str, Any] | None = None) \
            -> tuple[np.ndarray, dict[str, Any]]:
        """
        Defines the reset operation to be done after each episode.

        :param seed: seed for the random number generator
        :param options: Additional information to specify how the environment is reset
        :return: The current observation and additional information.
        """
        super().reset(seed=seed, options=options)
        self.steps_this_episode = 0

        self.update_values_from_scm_sample()

        return self.get_observation(), {}

    def render(self) -> NoReturn:
        """
        Renders the environment depending on the mode specified at initialization. When 'render_mode' is set to 'human',
        the causal graph is drawn with corresponding values. When 'render_mode' is set to 'terminal', the current values
        of the SCM's variables are printed.
        """
        if self.render_mode == 'human':
            self.scm.draw_graph()

        elif self.render_mode == 'terminal':
            print(self.scm.endogenous_vars, self.scm.exogenous_vars)

    def update_values_from_scm_sample(self) -> NoReturn:
        """Samples the next sample from the SCM and stores it in the appropriate buffers."""
        # get the next sample
        self.endogenous_vars, self.exogenous_vars = self.scm.get_next_sample()
        # add it to the history of samples
        self.samples_so_far.loc[len(self.samples_so_far)] = dict(self.endogenous_vars, **self.exogenous_vars)
        # determine the intervention targets and one-hot encode them
        targets = self.scm.get_intervention_targets()
        targets_row = {k: 1 if k in targets else 0 for k in self.endogenous_vars.keys()}
        self.targets_so_far.loc[len(self.targets_so_far)] = targets_row

    def clear_samples(self) -> NoReturn:
        """Clears buffer with samples collected so far."""
        self.samples_so_far = DataFrame(columns=list(self.scm.endogenous_vars.keys()) +
                                                list(self.scm.exogenous_vars.keys()))

    def clear_intervention_targets(self) -> NoReturn:
        """Clears buffer with intervention targets collected so far."""
        self.targets_so_far = DataFrame(columns=list(self.scm.endogenous_vars.keys()))

    def get_intervention_from_action(self, action: int) -> Tuple[str, Tuple[Callable, Dict]]:
        """Returns the intervention given the index in the discrete action space."""
        return self.possible_interventions[action]

    def get_causal_structure(self) -> nx.DiGraph:
        """
        Returns the causal graph of the SCM.

        :return: nx.DiGraph that represents the causal graph.
        """
        return self.scm.create_graph()

    def add_possible_intervention(self, intervention: Tuple[str, Tuple[Callable, Dict]]) -> NoReturn:
        """
        Adds an intervention to the action space. Caution: this operation implies adding actions to the action space.

        :param intervention: the intervention that should be added to the possible interventions. The first element of
        the tuple is the intervention target and the second element is the callable function representing the causal
        assignment together with a dictionary mapping the parameters of the callable to the names of the variables.
        """
        assert intervention[0] in self.scm.endogenous_vars, (f"The intervention target {intervention[0]} is not an "
                                                             f"endogenous variable in the SCM.")
        self.possible_interventions.append(intervention)
        self.action_space = spaces.Sequence(spaces.Discrete(len(self.possible_interventions)))

    # Task-specific methods. Override in inheriting class to adapt behaviour
    def get_observation(self) -> np.ndarray:
        """
        Determines the observation of the current step and returns it.

        :return: the observation as a vector.
        """
        return np.array(dict(self.endogenous_vars, **self.exogenous_vars).values())

    def determine_reward(self) -> float:
        """
        Calculates the reward for the current step and returns it.

        :return: the reward of the current step
        """
        return 0.0

    def determine_truncated(self) -> bool:
        """
        Determines whether the current epsisode ends because of a condition outside of the MDP definition.

        :return: whether the current episode ends.
        """
        return False

    def determine_terminated(self) -> bool:
        """
        Determines whether a terminal state was reached within the MDP definition.

        :return: whether the current episode ends.
        """
        return False




import CausalPlayground
from CausalPlayground import StructuralCausalModel
from CausalPlayground import SCMGenerator
import unittest
from functions import *
import random


class TestSCM(unittest.TestCase):
    def __init__(self, *args,  **kwargs):
        super(TestSCM, self).__init__(*args, **kwargs)
        self.scm = SCMGenerator(all_functions={'linear': f_linear}, seed=42).create_random(possible_functions=["linear"], n_endo=20, n_exo=0)[0]
        self.test_scm = StructuralCausalModel()
        self.test_scm.add_endogenous_var('A', lambda noise: noise + 5, {'noise': 'U'})
        self.test_scm.add_exogenous_var('U', random.randint, {'a': 3, 'b': 8})
        self.test_scm.add_endogenous_var('Effect', lambda x: x * 2, {'x': 'A'})

    def test_creation(self):
        sample = self.test_scm.get_next_sample()
        u = sample[1]['U']
        a = sample[0]['A']
        effect = sample[0]['EFFECT']
        self.assertIn(u, [3, 4, 5, 6, 7, 8])
        self.assertEqual(a, u+5)
        self.assertEqual(effect, a*2)

    def test_create_from_graph(self):
        graph = self.test_scm.create_graph()
        generator = SCMGenerator(all_functions={'linear': f_linear}, seed=1)
        scm = generator.create_scm_from_graph(graph=graph, possible_functions=['linear'],
                                              exo_distribution=random.randint, exo_distribution_kwargs={'a': 2, 'b': 5})
        self.assertTrue(scm.functions['A'][1] == {'U': 'U'})
        sample = scm.get_next_sample()
        u = sample[1]['U']
        a = sample[0]['A']
        effect = sample[0]['EFFECT']
        self.assertIn(u, [2, 3, 4, 5])
        self.assertTrue(effect <= 2*a)
    
    def test_intervention(self):
        # do an intervention and compare before and after
        x0 = self.scm.get_next_sample()[0]['X0']
        self.scm.do_interventions([("X0", (lambda: 5, {})), ("X1", (lambda x0: x0+1, {'x0': 'X0'}))])
        x0_do = self.scm.get_next_sample()[0]['X0']
        x1_do = self.scm.get_next_sample()[0]['X1']
        self.assertTrue(x0_do == 5)
        self.assertTrue(x1_do == 6)
        self.assertFalse(x0 == x0_do)

        # test graphical implications
        graph = self.scm.create_graph()
        # X1 should only have 1 parent
        self.assertEqual(len([e for e in graph.edges if e[1] == 'X1']), 1)
        # X0 should have no parents
        self.assertEqual(len([e for e in graph.edges if e[1] == 'X0']), 0)

        # test None - intervention
        self.scm.do_interventions([None, ("X1", (lambda x0: x0 + 1, {'x0': 'X0'}))])
        self.scm.undo_interventions()

    def test_undo_intervention(self):
        self.scm.do_interventions([("X0", (lambda: 5, {})), ("X1", (lambda x0: x0 + 1, {'x0': 'X0'}))])
        x1_do = self.scm.get_next_sample()[0]['X1']
        self.scm.undo_interventions()
        x1 = self.scm.get_next_sample()[0]['X1']
        self.assertFalse(x1_do == x1)

    def test_intervention_targets(self):
        """Are the correct intervention targets returned"""
        self.scm.do_interventions([("X0", (lambda: 5, {})), ("X1", (lambda x0: x0 + 1, {'x0': 'X0'}))])
        self.assertTrue(self.scm.get_intervention_targets() == ["X0", "X1"])
        self.scm.undo_interventions()
        self.assertTrue(self.scm.get_intervention_targets() == [])


class TestRandNN(unittest.TestCase):
    def __init__(self, *args,  **kwargs):
        super(TestRandNN, self).__init__(*args, **kwargs)
        self.scm1 = SCMGenerator(all_functions={'linear': f_linear}, seed=42).create_random(possible_functions=["linear"], n_endo=10, n_exo=0)[0]
        # self.scm2 = SCMGenerator(seed=42).create_random(possible_functions=["NN"], n_endo=6, n_exo=8,
        #                                                 exo_distribution=random.random, exo_distribution_kwargs={})[0]

    def test_sampling_types(self):
        vals = self.scm1.get_next_sample()[0]
        self.assertTrue(isinstance(vals['X0'], float))
        self.assertTrue(len(vals) == 10)

        # endo, exo = self.scm2.get_next_sample()
        # self.assertTrue(isinstance(endo[0], float))
        # self.assertTrue(len(endo) == 6)
        # self.assertTrue(isinstance(exo[0], float))
        # self.assertTrue(len(exo) == 8)

    def test_determinism(self):
        vals1 = self.scm1.get_next_sample()[0]
        vals2 = self.scm1.get_next_sample()[0]
        self.assertEqual(vals1, vals2)

        # vals1 = self.scm2.get_next_sample()[0]
        # vals2 = self.scm2.get_next_sample()[0]
        # self.assertNotEqual(vals1, vals2)


class TestExogenous(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(TestExogenous, self).__init__(*args, **kwargs)

        self.scm_confounded_controlled = StructuralCausalModel()

        self.generator = SCMGenerator(all_functions={'linear': f_linear}, seed=42)
        self.scm_unconfounded = self.generator.create_random(possible_functions=["linear"], n_endo=5, n_exo=4,
                                                             exo_distribution=random.random, exo_distribution_kwargs={},
                                                             allow_exo_confounders=False)[0]
        self.scm_confounded = self.generator.create_random(possible_functions=["linear"], n_endo=5, n_exo=7,
                                                           exo_distribution=random.random, exo_distribution_kwargs={},
                                                           allow_exo_confounders=True)[0]

    def test_confounded_create_graph(self):
        "does the graph generated from scm.create_graph for a scm with exo vars contain the exo vars"
        # unconfounded case
        graph_unconfounded = self.scm_unconfounded.create_graph()
        exo_edges = [e for e in graph_unconfounded.edges if e[0] in self.scm_unconfounded.exogenous_vars]
        self.assertFalse(len(exo_edges) == 0)
        self.assertTrue(len(set([e[0] for e in exo_edges])) == len([e[0] for e in exo_edges]))

        # confounded case
        graph_confounded = self.scm_confounded.create_graph()
        exo_edges = [e for e in graph_confounded.edges if e[0] in self.scm_confounded.exogenous_vars]
        self.assertFalse(len(exo_edges) == 0)
        self.assertFalse(len(set([e[0] for e in exo_edges])) == len([e[0] for e in exo_edges]))

    def test_sampling(self):
        "does the sampling work with exo vars"
        sample1 = self.scm_unconfounded.get_next_sample()
        sample2 = self.scm_unconfounded.get_next_sample()
        self.assertNotEqual(list(sample1[0].values()), list(sample2[0].values()))  # endogenous vars
        self.assertNotEqual(list(sample1[1].values()), list(sample2[1].values()))  # exoogenous vars


if __name__ == '__main__':
    unittest.main()



"""Classes for generating causal DAGs and sets of causal DAGs"""

from typing import Tuple, List, Dict, Callable
from scipy.special import comb
import copy
import random
import networkx as nx
from tqdm import tqdm
import pickle
import warnings
from CausalPlayground import StructuralCausalModel


class SCMGenerator:
    """Class for generating SCMs"""

    seed: int
    """Seed for the random number generator."""
    all_functions: Dict[str, Callable]
    """Dictionary of possible functions that can be generated in SCMs generated by this generator. The callable that
    is defined in this dictionary should return the function that is used in the SCM."""
    def __init__(self, all_functions: Dict[str, Callable], seed: int):
        """
        Constructor for the SCM generator.

        :param all_functions: dictionary containing possible functions that can be generated in SCMs generation.
        :param seed: the seed for the random number generator.
        """
        self.seed = seed
        self.all_functions = all_functions

    def create_random(self, possible_functions: List[str], n_endo: int, n_exo: int, exo_distribution: Callable = None,
                      exo_distribution_kwargs: dict = None, allow_exo_confounders: bool = False)\
            -> Tuple[StructuralCausalModel, set]:
        """
        Creates and returns a random StructualCausalModel by first generating a random graph and then randomly choosing
        a function for each edge in the graph based on the possible functions.

        :param possible_functions: list of function that can be used as causal relations. These should correspond to one
        of the strings defined in `self.all_functions`.
        :param n_endo: number of endogenous variables.
        :param n_exo: number of exogenous variables.
        :param exo_distribution: distribution of the exogenous variables. This distribution is applied to all
        exogenous variables.
        :param exo_distribution_kwargs: keyword arguments for the distribution of exogenous variables
        :param allow_exo_distribution: true if exogenous confounders should be generated.

        :return: the random scm and the set of edges that have been removed
        """
        if n_exo > 0:
            assert exo_distribution, 'please provide a distribution for your exogenous variables'
            assert not exo_distribution_kwargs == None, 'please provide kwargs for the distribution of your exogenous variables'
        random.seed(self.seed)
        graph_generator = CausalGraphGenerator(n_endo, n_exo, allow_exo_confounders, seed=self.seed)
        graph, removed_edges = graph_generator.generate_random_graph()
        self.seed += 1
        return self.create_scm_from_graph(graph, possible_functions, exo_distribution, exo_distribution_kwargs),\
               removed_edges

    def create_scm_from_graph(self, graph: nx.DiGraph, possible_functions: List[str], exo_distribution: Callable,
                              exo_distribution_kwargs: dict) -> StructuralCausalModel:
        """
        Generates an SCM following the structure of a given directed graph. A function is randomly chosen from
        `possible_functions` for each edge in the graph.

        :param graph: causal graph
        :param possible_functions: list of function that can be used as causal relations. These should correspond to one
        of the strings defined in 'self.all_functions'
        :param exo_distribution: distribution of the exogenous variables. This distribution is applied to all
        exogenous variables.
        :param exo_distribution_kwargs: keyword arguments for the distribution of exogenous variables
        """
        scm = StructuralCausalModel()
        random.seed(self.seed)

        for n in graph.nodes:
            parents = [p for p in graph.predecessors(n)]
            if graph.nodes[n]['type'] == 'endo':
                # randomly choose one of the possible functions for the current node
                current_function = random.choice(possible_functions)
                scm.add_endogenous_var(n, self.all_functions[current_function](parents), {p: p for p in parents})
            else:
                scm.add_exogenous_var(n, exo_distribution, exo_distribution_kwargs)

        self.seed += 1
        return scm


class CausalGraphGenerator:
    """Class for generating random directed acyclic graphs."""

    allow_exo_confounders: bool
    """flag whether the generated graphs can have exogenous confounders. If False, there will be at most 1 exogenous 
    variable per endogenous variable."""
    seed: int
    """Seed for the random number generator."""
    endo_vars: List[str]
    """List of variable names of the endogenous variables. For the automatic generation in this class, endogenous 
    variables will be named 'Xn' with n ascending from 0."""
    exo_vars: List[str]
    """List of variable names of the exogenous variables. For the automatic generation in this class, exogenous 
    variables will be named 'Un' with n ascending from 0."""
    potential_causes: Dict[str, List[str]]
    """Dictionary of potential causes for each variable in the graph. This helps to manage exogenous variables and their
    confounding effects."""
    fully_connected_graph: nx.DiGraph
    """Fully connected graph of the endogenous variables where every endogenous node is connected to every other 
    endogenous node. The edges from exogenous variables depend on the 'allow_exogenous_confounders' flag."""
    def __init__(self, n_endo: int, n_exo: int = 0, allow_exo_confounders: bool = False, seed: int = 42):
        """
        Constructor defining parameters for the randomly generated graphs.

        :param n_endo: number of endogenous variables.
        :param n_exo: number of exogenous variables.
        :param allow_exo_confounders: flag whether the generated graphs can have exogenous confounders. If False, there
        will be at most 1 exogenous variable per endogenous variable.
        :param seed: seed for the random number generator.
        """
        self.allow_exo_confounders = allow_exo_confounders
        self.seed = seed

        # create var names
        self.endo_vars = ['X' + str(i) for i in range(n_endo)]
        self.exo_vars = ['U' + str(i) for i in range(n_exo)]

        # determine potential causes for each endogenous var
        self.potential_causes = {}
        exo_copy = copy.deepcopy(self.exo_vars)
        for v in self.endo_vars:
            if allow_exo_confounders:  # if confounding is allowed, any node can be the parent of an endogenous var
                self.potential_causes[v] = self.endo_vars + self.exo_vars
            else:  # else, only one exo var can be the cause of an endogenous var
                if not len(exo_copy) == 0:
                    self.potential_causes[v] = self.endo_vars + [exo_copy.pop()]
                else:
                    self.potential_causes[v] = self.endo_vars + []
            # disable self-cycles
            self.potential_causes[v].remove(v)

        del exo_copy
        # create a fully-connected graph based on the confounder settings and store it for later generation
        self.fully_connected_graph = self._make_fully_connected_dag()

    def _make_fully_connected_dag(self) -> nx.DiGraph:
        """
        Helper function that creates and returns a fully connected, possibly cyclic graph. In this graph the exogenous
        variables are roots.

        :return: A fully connected, directed graph.
        """
        graph = nx.DiGraph()
        [graph.add_node(u, type='exo') for u in self.exo_vars]
        [graph.add_node(v, type='endo') for v in self.endo_vars]
        for n, causes in self.potential_causes.items():
            [graph.add_edge(c, n) for c in causes]
        return graph

    def generate_random_graph(self, method: str = None, p: float = 0.3) -> Tuple[nx.DiGraph, set[Tuple[str, str]]]:
        """
        Creates and returns a random nx.DiGraph either by first creating a fully connected graph or creating an Edos
        Reny graph and then randomly deleting one edge after the other until it is acyclic.

        When generating Edos Reny graphs, exogenous variables are not supported for generation. (You can always add
        additional nodes and edges after generation).

        :param method: either 'ER' to generate an Edos Reny graph, or any other string for creating a graph based on
                        random deletion of edges from a fully-connected graph
        :param p: probability of an edge being added to the graph when using the ER method.
        :return: the random nx.DiGraph and the set of edges that have been removed from the fully-connected graph.
        """
        if method == 'ER':  # create erdos reny graph
            assert len(self.exo_vars) == 0, "exogenous variables are not supported when generating ER graphs"
            graph, removed_edges = self._create_er_graph(p)
        else:
            graph, removed_edges = self._create_from_fully_connected()
        self.seed += 1
        return graph, removed_edges

    def _create_er_graph(self, p: float = 0.1) -> (nx.DiGraph, set[Tuple[str, str]]):
        """
        Helper method that creates an Edos Reny graph and then deletes edges until the graph is acyclic.

        :param p: probability of an edge being added to the graph.
        :return: the acyclic nx.DiGraph and the set of edges that have been removed from the fully-connected graph.
        """
        graph = nx.fast_gnp_random_graph(len(self.endo_vars), p=p, seed=self.seed, directed=True)

        # rename the nodes and types according to the fully connected graph
        graph = nx.relabel_nodes(graph, {n: m for (n, m) in zip(graph.nodes, self.fully_connected_graph.nodes)})
        nx.set_node_attributes(graph,
                               {n: {'type': self.fully_connected_graph.nodes()[m]['type']} for (n, m) in
                                zip(graph.nodes, self.fully_connected_graph.nodes)})

        # make acyclic
        while not nx.is_directed_acyclic_graph(graph):
            random_edge = random.choice(list(graph.edges()))
            graph.remove_edge(random_edge[0], random_edge[1])

        # compute which edges are removed
        removed_edges = nx.difference(self.fully_connected_graph, graph).edges

        return graph, set(removed_edges)

    def _create_from_fully_connected(self) -> (nx.DiGraph, set[Tuple[str, str]]):
        """
        Helper method that deletes edges from the fully-connected graph until it is acyclic.

        :return: the acyclic nx.DiGraph and the set of edges that have been removed from the fully-connected graph.
        """
        # generate fully connected graph
        graph = copy.deepcopy(self.fully_connected_graph)
        # delete random edges from the endogenous subgraph until acyclic
        removed_edges = set()
        random.seed(self.seed)
        # determine outgoing edges from exogenous variables
        endogenous_edges = list(graph.edges)
        if not self.allow_exo_confounders:  # only delete the exogenous edges if we allow confounding
            [endogenous_edges.remove(e) for e in graph.edges if e[0] in self.exo_vars]

        # delete edges until the graph is acyclic
        while not nx.is_directed_acyclic_graph(graph):
            random_edge = random.sample(endogenous_edges, 1)[0]
            removed_edges.add(random_edge)
            endogenous_edges.remove(random_edge)
            graph.remove_edge(random_edge[0], random_edge[1])
        return graph, set(removed_edges)


class CausalGraphSetGenerator:
    """
    Represents a collection of unique DAGs and facilitates their bookkeeping.
    """

    generator: CausalGraphGenerator
    """Object generating the causal graphs."""
    graphs: List[nx.DiGraph]
    """List containing all generated graphs."""
    max_endo_dags: int
    """Number of all possible DAGs with 'n_endo' nodes."""
    def __init__(self, n_endo: int, n_exo: int = 0, allow_exo_confounders: bool = False, seed: int = 42):
        """
        Constructor to parametrize the CausalGraphSetGenerator.

        :param n_endo: number of endogenous variables.
        :param n_exo: number of exogenous variables.
        :param allow_exo_confounders: flag whether the generated graphs can have exogenous confounders. If False, there
        will be at most 1 exogenous variable per endogenous variable.
        :param seed: seed for the random number generator.
        """
        self.generator = CausalGraphGenerator(n_endo, n_exo=n_exo,
                                              allow_exo_confounders=allow_exo_confounders,
                                              seed=seed)
        self.graphs = []
        self.max_endo_dags = CausalGraphSetGenerator.max_n_dags(n_endo)

    def generate(self, n: int, method: str = None, p: float = 0.1):
        """
        Generate n distinct causal DAGs. The newly generated graphs are only added to the set of graphs, if the edges
        that have been removed during the generation are different from all other graphs in the set. This ensures the
        uniqueness of the generated, random graphs.

        :param n: how many DAGs to create
        :param method: which method to use to generate the graphs.
        :param p: probability with which an edge is added if ER is used.
        """
        # check whether more DAGs are to be created then combinatorically possible. Only do this if n is not
        # too big because computation takes forever for n > 20 and for such values there exist over 2.3e+72
        # different graphs
        n_exo = len(self.generator.exo_vars)
        if n > self.max_endo_dags * (n_exo + 1):
            n = self.max_endo_dags * (n_exo + 1)
            warnings.warn('Only ', self.max_endo_dags * (n_exo + 1), ' graphs can be created.')

        self.graphs = []
        rem_edges_list = []
        resampled = 0
        print('Creating graphs...')
        pbar = tqdm(total=n - 1)
        while len(self.graphs) < n - 1:
            graph, rem_edges = self.generator.generate_random_graph(method=method, p=p)
            if any([rem_edges == other for other in rem_edges_list]):
                resampled += 1
                continue
            else:
                self.graphs.append(graph)
                rem_edges_list.append(rem_edges)
                pbar.update(1)
        pbar.close()
        print(resampled, 'models resampled')

    def save(self, filepath: str, mode: str = 'wb'):
        """
        Save the generated graphs to a file.

        :param filepath: path to the file.
        :param mode: mode in which to open the file.
        """
        with open(str(filepath), str(mode)) as f:
            pickle.dump(self.graphs, f)

    def load(self, filepath: str):
        """
        Load graphs from files. Warning: does not check whether the loaded graphs are unique.

        :param filepath: path to the file containing the graphs.
        """
        # Warning: doesn't check whether the loaded graphs are actually a set
        with open(str(filepath), 'rb') as f:
            graphs = pickle.load(f)
            if not isinstance(graphs[0], nx.DiGraph):
                raise TypeError('The file does not contain nx.DiGraphs')
            else:
                self.graphs = graphs

    @staticmethod
    def max_n_dags(n_vertices: int) -> int:
        """
        Computes the maximal number of different DAGs over n_vertices nodes. Implemented as in Robinson (1973)

        :param n_vertices: number of nodes in the DAG.
        :return: max number of possible DAGs.
        """
        if n_vertices < 0:
            return 0
        elif n_vertices == 0:
            return 1
        else:
            summ = 0
            for k in range(1, n_vertices + 1):
                summ += (-1) ** (k - 1) * comb(n_vertices, k) * 2 ** (
                        k * (n_vertices - k)) * CausalGraphSetGenerator.max_n_dags(n_vertices - k)
            return int(summ)
